// Written by Tilman Neumann (www.tilman-neumann.de) 2016, 2017
// Any commercial use of this software requires the written consent of the author.
// Usage and republication are free for non-commercial purposes, as long as this header is carried along.
// The author takes no warranties for the correctness of the software.
package de.tilman_neumann.math.factor.basics;

import org.apache.log4j.Logger;

/**
 * A reused buffer to store big factors of partials temporarily during trial division.
 */
// TODO: Overpowered for arrays having at most 2 (QS) or 3 (CFrac) elements
public class SortedLongArray {
	@SuppressWarnings("unused")
	private static final Logger LOG = Logger.getLogger(SortedLongArray.class);
	
	private long[] factors;
	private byte[] exponents;
	private int size;
	
	public SortedLongArray() {
		// we will hardly get more than 3 large factors, but this is a re-used bufffer, so size 10 should match
		// any possible requirements and is no uhge waste of memory.
		this.factors = new long[10];
		this.exponents = new byte[10];
	}
	
	/**
	 * reset() must be called before using for a new Q.
	 */
	public void reset() {
		this.size = 0;
	}
	
	/**
	 * Add a factor.
	 * @param factor
	 * @param power
	 */
	public void add(long factor) {
		//LOG.debug("add small factor " + factor);
		if (size>0) {
			// find position: for short lists a linear search is ok.
			int insertPosition = 0;
			for (; insertPosition<size; insertPosition++) {
				if (factor <= factors[insertPosition]) break;
			}
			if (insertPosition<size) {
				if (factor == factors[insertPosition]) {
					// the factor exists already -> just increment exponent
					exponents[insertPosition]++;
					return;
				}
				// the new factor that must be inserted before some other content.
				// -> shift the elements bigger than factor:
				for (int i=size-1; i>=insertPosition; i--) {
					factors[i+1] = factors[i];
					exponents[i+1] = exponents[i];
				}
				// and add the new factor
				factors[insertPosition] = factor;
				exponents[insertPosition] = 1;
				size++;
				return;
			}
			// insertPosition==size -> append new last element
			factors[size] = factor;
			exponents[size] = 1;
			size++;
			return;
		}
		// size is still 0
		factors[0] = factor;
		exponents[0] = 1;
		size = 1;
	}
	
	/**
	 * Add a factor to the given power.
	 * @param factor
	 * @param power
	 */
	public void add(long factor, byte power) {
		//LOG.debug("add small factor " + factor);
		if (size>0) {
			// find position: for short lists a linear search is ok.
			int insertPosition = 0;
			for (; insertPosition<size; insertPosition++) {
				if (factor <= factors[insertPosition]) break;
			}
			if (insertPosition<size) {
				if (factor == factors[insertPosition]) {
					// the factor exists already -> just increment exponent
					exponents[insertPosition] += power;
					return;
				}
				// the new factor that must be inserted before some other content.
				// -> shift the elements bigger than factor:
				for (int i=size-1; i>=insertPosition; i--) {
					factors[i+1] = factors[i];
					exponents[i+1] = exponents[i];
				}
				// and add the new factor
				factors[insertPosition] = factor;
				exponents[insertPosition] = power;
				size++;
				return;
			}
			// insertPosition==size -> append new last element
			factors[size] = factor;
			exponents[size] = power;
			size++;
			return;
		}
		// size is still 0
		factors[0] = factor;
		exponents[0] = power;
		size = 1;
	}
	
	/**
	 * @param i
	 * @return the i.th large factor entry
	 */
	public long get(int i) {
		return factors[i];
	}
	
	public int size() {
		return size;
	}
	
	public long[] copyFactors() {
		long[] copy = new long[size];
		System.arraycopy(factors, 0, copy, 0, size);
		return copy;
	}
	
	public byte[] copyExponents() {
		byte[] copy = new byte[size];
		System.arraycopy(exponents, 0, copy, 0, size);
		return copy;
	}
	
	public String toString() {
		if (size==0) return "(empty)";
		String str = "";
		for (int i=0; i<size; i++) {
			str += factors[i] + "^" + exponents[i] + " * ";
		}
		return str.substring(0, str.length()-3);
	}
}
