// Written by Tilman Neumann (www.tilman-neumann.de) 2016, 2017
// Any commercial use of this software requires the written consent of the author.
// Usage and republication are free for non-commercial purposes, as long as this header is carried along.
// The author takes no warranties for the correctness of the software.
package de.tilman_neumann.math.factor.cfrac;

import java.math.BigInteger;
import java.util.HashSet;

import org.apache.log4j.Logger;

import de.tilman_neumann.math.base.bigint.sequence.IntegerSequence;
import de.tilman_neumann.math.factor.FactorAlgorithmBase;
import de.tilman_neumann.math.factor.FactorException;
import de.tilman_neumann.math.factor.basics.congruence.AQPair;
import de.tilman_neumann.math.factor.basics.congruence.CongruenceCollector;
import de.tilman_neumann.math.factor.basics.matrixSolver.FactorTest;
import de.tilman_neumann.math.factor.basics.matrixSolver.FactorTest01;
import de.tilman_neumann.math.factor.basics.matrixSolver.MatrixSolver;
import de.tilman_neumann.math.factor.basics.matrixSolver.SmoothSolverController;
import de.tilman_neumann.math.factor.basics.primeBase.PrimeBaseBuilder;
import de.tilman_neumann.math.factor.basics.primeBase.PrimeBaseBuilder02;
import de.tilman_neumann.math.factor.cfrac.tdiv.TDiv_CF63;

import static de.tilman_neumann.math.base.bigint.BigIntConstants.*;
import static org.junit.Assert.*;

/**
 * CFrac = Shanks' SQUFOF algorithm + carry along continuant recurrence + collect smooth relations + LinAlg solver.<br/>
 * 
 * Stopping criterion for a single k: after complete period or at a maximum number of iterations.
 * 
 * CFrac63:
 * P- and Q-values are constrained to long size.
 * This implies that big factors fit into int.
 */
// Choice of parameters:
// * The best CFrac63 for N < 80 bits is with C=0.275, maxSuSmoothExp~0.15+bits/1000, auxA, cc01.
// * (stopRoot, stopMult) has no influence if it is big enough, because then a second k is hardly ever needed. (5, 1.5) is ok.
public class CFrac63 extends FactorAlgorithmBase {
	private static final Logger LOG = Logger.getLogger(CFrac63.class);
	private static final boolean DEBUG = false;

	// input
	private BigInteger N, kN;
	private long floor_sqrt_kN;
	private IntegerSequence<BigInteger> kSequence;
	/** Test all Q or only those Q_i+1 with odd i ? "All Q" boosts performance for N > 45 bits approximately */
	private boolean use_all_i; // 

	// maximum number of SquFoF iterations for a single k
	private int stopRoot;
	private float stopMult;
	private long maxI;
	
	// prime base
	private float C;
	private int primeBaseSize;
	private PrimeBaseBuilder primeBaseBuilder = new PrimeBaseBuilder02();
	private int[] primesArray;
	/** The union of all reduced prime bases -- this is the number of variables in the equation system */
	private HashSet<Integer> combinedPrimesSet;

	// factorizer for Q
	private TDiv_CF63 auxFactorizer;
	private float maxSufficientSmoothExp;
	
	// collects the congruences we find
	private CongruenceCollector congruenceCollector;
	
	// the number of congruences we need to find before we try to solve the smooth congruence equation system
	private int requiredSmoothCongruenceCount;
	// extra congruences to have a bigger chance that the equation system solves. the likelihood is >= 1-2^(extraCongruences+1)
	private int extraCongruences;
	/** A controller for the solver used for smooth congruence equation systems. */
	private SmoothSolverController smoothSolverController;

	// time capturing
	private long startTime, linAlgStartTime;
	
	/**
	 * Standard constructor.
	 * @param primeTestBits
	 * @param kSequence multiplier generator
	 * @param use_all_i
	 * @param stopRoot order of the root to compute the maximum number of iterations
	 * @param stopMult multiplier to compute the maximum number of iterations
	 * @param C multiplier for prime base size
	 * @param maxSufficientSmoothExp
	 * @param auxFactorizer the algorithm to find smooth Q
	 * @param extraCongruences the number of surplus congruences we collect to have a greater chance that the equation system solves.
	 * @param matrixSolver matrix solver for the smooth congruence equation system
	 */
	public CFrac63(int primeTestBits, IntegerSequence<BigInteger> kSequence, boolean use_all_i, int stopRoot, float stopMult, float C,
				   float maxSufficientSmoothExp, TDiv_CF63 auxFactorizer, int extraCongruences, MatrixSolver<Integer> matrixSolver) {
		super(primeTestBits);
		this.kSequence = kSequence;
		this.use_all_i = use_all_i;
		this.stopRoot = stopRoot;
		this.stopMult = stopMult;
		this.C = C;
		this.maxSufficientSmoothExp = maxSufficientSmoothExp;
		this.auxFactorizer = auxFactorizer;
		this.congruenceCollector = new CongruenceCollector();
		this.extraCongruences = extraCongruences;
		this.smoothSolverController = new SmoothSolverController(matrixSolver);
	}

	@Override
	public String getName() {
		return "CFrac63(" + kSequence.getName() + ", all_i=" + use_all_i + ", stop=(" + stopRoot + ", " + stopMult + "), C=" + C + ", maxSuSmoothExp=" + maxSufficientSmoothExp + ", " + auxFactorizer.getName() + ")";
	}
	
	/**
	 * Test the current N.
	 * @return factor, or null if no factor was found.
	 */
	public BigInteger findSingleFactor(BigInteger N) {
		this.N = N;
		this.startTime = System.currentTimeMillis();

		// compute prime base size
		double N_dbl = N.doubleValue();
		double lnN = Math.log(N_dbl);
		double lnlnN = Math.log(lnN);
		double lnNPow = 0.5; // heuristics for quadratic sieve
		// we want that the exponents of lnN and lnlnN sum to 1
		this.primeBaseSize = (int) (Math.exp(Math.pow(lnN, lnNPow) * Math.pow(lnlnN, 1-lnNPow) * C));
		if (DEBUG) LOG.debug("N = " + N + ": primeBaseSize = " + primeBaseSize);
		this.primesArray = new int[primeBaseSize];

		// compute the biggest unfactored rest where some Q is still considered "sufficiently smooth".
		double maxSufficientSmoothRest = Math.pow(N_dbl, maxSufficientSmoothExp);
		
		// initialize sub-algorithms for N
		this.auxFactorizer.initialize(N, maxSufficientSmoothRest);
		FactorTest factorTest = new FactorTest01(N);
		this.congruenceCollector.initialize(N, factorTest, false);
		this.combinedPrimesSet = new HashSet<Integer>();
		this.smoothSolverController.initialize(N, factorTest);

		// max iterations: there is no need to account for k, because expansions of smooth kN are typically not longer than those for N.
		// long maxI is sufficient to hold any practicable number of iteration steps (~9.22*10^18);
		// stopRoot must be chosen appropriately such that there is no overflow of long values.
		// with stopRoot=5, the overflow of longs starts at N~6.67 * 10^94...
		this.maxI = (long) (stopMult*Math.pow(N_dbl, 1.0/stopRoot));
		// for each new N start again with the first k
		kSequence.reset(N);
		while (true) {
			// get a new k, return immediately if kN is square
			BigInteger k = kSequence.next();
			this.kN = k.multiply(N);
			floor_sqrt_kN = (long) Math.sqrt(kN.doubleValue()); // faster than BigInteger sqrt; but the type conversion may give ceil(sqrt(kN)) for some N >= 54 bit
			BigInteger sqrt_kN_big = BigInteger.valueOf(floor_sqrt_kN);
			long diff = kN.subtract(sqrt_kN_big.multiply(sqrt_kN_big)).longValue();
			if (diff==0) return N.gcd(sqrt_kN_big);
			if (diff<0) {
				// floor_sqrt_kN was too big, diff too small -> compute correction
				floor_sqrt_kN--;
				diff += (floor_sqrt_kN<<1) + 1;
			}

			// Create the reduced prime base for kN:
			primeBaseBuilder.computeReducedPrimeBase(kN, primeBaseSize, primesArray);
			// add new reduced prime base to the combined prime base
			for (int i=0; i<primeBaseSize; i++) combinedPrimesSet.add(primesArray[i]);
			// we want: #equations = #variables + some extra congruences
			this.requiredSmoothCongruenceCount = combinedPrimesSet.size() + extraCongruences;

			try {
				// initialize the Q-factorizer with new prime base
				this.auxFactorizer.initialize(kN, primeBaseSize, primesArray); // may throw FactorException
				// search square Q_i
				test(diff);
			} catch (FactorException fe) {
				long endTime = System.currentTimeMillis();
				if (DEBUG) LOG.info("Found factor of N=" + N + " in " + (endTime-startTime) + "ms (LinAlgPhase took " + (endTime-linAlgStartTime) + "ms)");
				return fe.getFactor();
			}
		}
	}

	protected void test(long Q_ip1) throws FactorException {
		// initialization for first iteration step
		long i = 0;
		BigInteger A_im2 = null;
		BigInteger A_im1 = ONE;
		BigInteger A_i = BigInteger.valueOf(floor_sqrt_kN);
		long P_im1 = 1;
		long P_i = floor_sqrt_kN;
		long Q_i = 1;
		
		// first iteration step
		long two_floor_sqrt_kN = floor_sqrt_kN<<1;
		while (true) {
			if (DEBUG) verifyCongruence(i, A_i, BigInteger.valueOf(Q_ip1));
			// [McMath 2004] points out (on SquFoF) that we have to look for square Q_i at some even i.
			// Here I test Q_i+1, so I have to look for square Q_i+1 at odd i.
			// In CFRAC, square congruences are also tested in the CongruenceCollector,
			// but doing it here before trial division is good for performance.
			boolean isSquare = false;
			if (i%2==1) {
				long Q_ip1_sqrt = (long) Math.sqrt(Q_ip1);
				if (Q_ip1_sqrt*Q_ip1_sqrt==Q_ip1) {
					// Q_i+1 is square -> test gcd
					BigInteger gcd = N.gcd(A_i.subtract(BigInteger.valueOf(Q_ip1_sqrt)));
					if (gcd.compareTo(ONE)>0 && gcd.compareTo(N)<0) throw new FactorException(gcd);
					isSquare = true;
				}
			}
			if (isSquare==false && (use_all_i || i%2==1)) {
				// Q_i+1 is not square and the i is right, too -> check for smooth relations.
				// Here a constraint on the size of Q would mean a severe performance penalty!
				long Q_test = i%2==1 ? Q_ip1 : -Q_ip1; // make Q congruent A^2
				AQPair aqPair = auxFactorizer.test(A_i, Q_test);
				if (DEBUG) LOG.debug("N = " + N + ": Q_test = " + Q_test + " -> aqPair = " + aqPair);
				if (aqPair!=null) {
					// the Q was sufficiently smooth
					linAlgStartTime = System.currentTimeMillis(); // just in case it really starts
					boolean addedSmooth = congruenceCollector.add(aqPair);
					if (addedSmooth) {
						int smoothCongruenceCount = congruenceCollector.getSmoothCongruenceCount();
						if (smoothCongruenceCount >= requiredSmoothCongruenceCount) {
							// try to solve equation system
							//LOG.debug("#smooths = " + smoothCongruenceCount + ", #requiredSmooths = " + requiredSmoothCongruenceCount);
							smoothSolverController.solve(congruenceCollector.getSmoothCongruences()); // throws FactorException
							// no factor exception -> extend equation system and continue searching smooth congruences
							requiredSmoothCongruenceCount += extraCongruences;
						}
					}
				}
			}
		
			// exit loop ?
			if (++i==maxI) return;
			
			// keep values from last round
			A_im2 = A_im1;
			A_im1 = A_i;
			P_im1 = P_i;
			long Q_im1 = Q_i;
			Q_i = Q_ip1;
			// compute next values
			long b_i = (floor_sqrt_kN + P_im1)/Q_i; // floor(rational result)
			P_i = b_i*Q_i - P_im1;
			Q_ip1 = Q_im1 + b_i*(P_im1-P_i);
			// carry along A_i % N from continuant recurrence
			A_i = addModN(mulModN(b_i, A_im1), A_im2);
			
			// stop when continuant period is complete
			if (b_i==two_floor_sqrt_kN) return;
		}
	}

	/**
	 * Verify congruence A_i^2 == (-1)^(i+1)*Q_i+1 (mod N)
	 * @param i
	 * @param A_i
	 * @param Q_ip1
	 */
	private void verifyCongruence(long i, BigInteger A_i, BigInteger Q_ip1) {
		assertTrue(Q_ip1.signum()>=0);
		// verify congruence A^2 == Q (mod N)
		BigInteger Q_test = i%2==1 ? Q_ip1 : Q_ip1.negate().mod(N);
		BigInteger div[] = A_i.pow(2).subtract(Q_test).divideAndRemainder(N);
		assertEquals(ZERO, div[1]); // works
		LOG.debug("A^2-Q = " + div[0] + " * N");
		LOG.debug("A^2 % N = " + A_i.pow(2).mod(N) + ", Q = " + Q_test);
		assertEquals(Q_test, A_i.pow(2).mod(N)); // works
	}
	
	/**
	 * Addition modulo N, with <code>a, b < N</code>.
	 * @param a
	 * @param b
	 * @return (a+b) mod N
	 */
	private BigInteger addModN(BigInteger a, BigInteger b) {
		BigInteger sum = a.add(b);
		return sum.compareTo(N)<0 ? sum : sum.subtract(N);
	}

	/**
	 * Multiplication (m*a) modulo N, with m often small and <code>a < N</code>.
	 * @param m
	 * @param a
	 * @return (m*a) mod N
	 */
	private BigInteger mulModN(long m_long, BigInteger a) {
		if (m_long<4) { // 0, 1, 10, 11
			int m = (int) m_long;
			switch (m) {
			case 0: return ZERO;
			case 1: return a;
			case 2: {
				BigInteger two_a = a.shiftLeft(1); // faster than 2*a or a+a
        		return two_a.compareTo(N)<0 ? two_a : two_a.subtract(N);
			}
			case 3: {
				BigInteger ma_modN = a.shiftLeft(1).add(a); // < 3*N
        		if (ma_modN.compareTo(N)<0) return ma_modN;
        		ma_modN = ma_modN.subtract(N); // < 2*N
        		return ma_modN.compareTo(N)<0 ? ma_modN : ma_modN.subtract(N);
			}
			}
			// adding case 4 does not help because then bitLength() does not fit exactly
		}
		BigInteger product = a.multiply(BigInteger.valueOf(m_long));
		return product.compareTo(N)<0 ? product : product.mod(N);
	}
}
